#if HAVE_CONFIG_H
# include "config.h"
#endif

#if HAVE_SYS_TYPES_H
# include <sys/types.h>
#endif

#if HAVE_SYS_SOCKET_H
# include <sys/socket.h>
#endif

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#if HAVE_NETDB_H
# include <netdb.h>
#endif

#include "csock.h"

/*
	Notes
	- gethostbyname is not reentrant, remember this when forking
	- checks wether host is an ipv4 address before resolving (if inet_aton is present)
*/

int create_socket(char *host, uint16_t port, int *sk)
{
	struct sockaddr_in sa;
	struct hostent *ip;
	int s, ret;
	
	s = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (!s) { return CSOCK_E_SOCKET; }
	memset(&sa, 0, sizeof(sa));
	sa.sin_port = htons(port);
	sa.sin_family = AF_INET;
#if HAVE_INET_ATON
	if (inet_aton(host, &sa.sin_addr.s_addr) == 0) {
#endif
		if (!(ip = gethostbyname(host))) { 
			ret = CSOCK_E_HOST;
			goto bad_end;
		}
		if (ip->h_length != sizeof(sa.sin_addr.s_addr)) {
			ret = CSOCK_E_LEN_MISMATCH;
			goto bad_end;
		}
		memcpy(&sa.sin_addr.s_addr, ip->h_addr, ip->h_length);
#ifdef HAVE_INET_ATON
	}
#endif
	ret = connect(s, (struct sockaddr *)&sa, sizeof(sa));
	if (ret == -1) {
		ret = CSOCK_E_CONNECT;
		goto bad_end;
	}
	*sk = s;
	return 0;
bad_end:
	close(s);
	return ret;
}

