#include <stdio.h>

#if HAVE_CONFIG_H
# include "config.h"
#endif

	/* standard C headers */
#if STDC_HEADERS
# include <string.h>
#else
# if !HAVE_STRCHR
#  define strchr index
#  define strrchr rindex
# endif
char *strchr (), *strrchr ();
# if !HAVE_MEMCPY
#  define memcpy(d, s, n) bcopy ((s), (d), (n))
#  define memmove(d, s, n) bcopy ((s), (d), (n))
# endif
#endif

	/* types & unistd */
#if HAVE_UNISTD_H
# if HAVE_SYS_TYPES_H
#  include <sys/types.h>
# endif
# include <unistd.h>
#endif

#if TIME_WITH_SYS_TIME 
#include <sys/time.h>
#include <sys/types.h>
#else
# if HAVE_SYS_TIME_H
#  include <sys/time.h>
# else
#  include <time.h>
# endif
#endif

	/* sys-wait */
#if HAVE_SYS_WAIT_H
# include <sys/wait.h>
#endif
#ifndef WEXITSTATUS
# define WEXITSTATUS(stat_val) ((unsigned)(stat_val) >> 8)
#endif
#ifndef WIFEXITED
# define WIFEXITED(stat_val) (((stat_val) & 255) == 0)
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_LIBGEN_H
# include <libgen.h>
#endif

#if HAVE_SYS_STAT_H
# include <sys/stat.h>
#endif

#if HAVE_SYSLOG_H
# include <syslog.h>
#endif

#if HAVE_PWD_H
# include <pwd.h>
#endif
#if HAVE_GRP_H
# include <grp.h>
#endif

#if HAVE_OPENSSL_SSL_H
# include <openssl/ssl.h>
#endif
#if HAVE_OPENSSL_ERR_H
# include <openssl/err.h>
#endif

#include <signal.h>

#include "sslhelp.h"
#include "tun.h"
#include "sptc.h"

#if !HAVE_WORKING_FORK
# define vfork fork
#endif

/* GLOBALS */
int 	child_pid;
tunnel 	t;
SSL_CTX *t_ctx;
FILE 	*errlog;

/* OPTIONS */
int select_to;
int inactivity_to;
char *pid_file;
char *cert_path;
	/* user and group to run as */
int sptc_uid = -1;
int sptc_gid = -1;

/* FUNCTIONS */
void usage(char *p)
{
#if VERIFY_CERT
# define CERT_OPT	"[-c certs] " 
#else
# define CERT_OPT	""
#endif
	
	fprintf(stderr, "usage: %s %s[-i inactivity timeout] "
			"[-p pid_file]  "
			"[-t select timeout] [-u username] "
			"[-g group] "
			"[-v] [-w working_dir] -- host port command "
			"arg1 ... argN\n", p, CERT_OPT);
	exit(E_SPTC_USAGE);
}

void cleanup(void)
{
	int status;
	tunnel_down(&t);
	waitpid(child_pid, &status, 0);	/* wait for child to exit */
	unlink(pid_file);
	syslog(SPTC_PRI, "clean exit (child-exit-code=%d)", 
			WEXITSTATUS(status));
	exit(WEXITSTATUS(status));	/* exit with status of child */
}

int write_pid(char *pid_file, int pid)
{
	FILE *pid_fd;
	
	pid_fd = fopen(pid_file, "w");
	if (!pid_fd)
		return 0;
	if (!fprintf(pid_fd, "%d\n", pid)) {
		fclose(pid_fd);
		return 0;
	}
	fclose(pid_fd);
	return 1;
}

void pipe_bus(int s, siginfo_t *si, void *bleh)
{
	qerr("caught PIPE");
		/* assume tunnel is down */
	tunnel_down(&t);
}

	/* catch SIGTERM and propogate to child */
void catch_term(int s, siginfo_t *si, void *bleh)
{
	qerr("caught TERM");
	kill(child_pid, SIGTERM);
	cleanup();
}

int spawn_command(char *p, char **args, int in_fd, int out_fd)
{
	extern char **environ;
	
	if (dup2(in_fd, fileno(stdin)) == -1)
		return E_SPTC_DUP2;
	if (dup2(out_fd, fileno(stdout)) == -1)
		return E_SPTC_DUP2;
	if (dup2(out_fd, fileno(stderr)) == -1)
		return E_SPTC_DUP2;
	
	execvp(p, args);
		/* should not get here if cmd is valid */
	return 0;
}

int copy_fd2tun(int src, tunnel *dst)
{
	char line[CPBUF_SIZE_MAX];
	int len;

	len = read(src, line, CPBUF_SIZE_MAX);
	if (len > 0) {
		return tunnel_write(dst, line, len);
	}
	else {
		return SPTC_CMD_EOF;
	}
}

int copy_tun2fd(tunnel *src, int dst)
{
	char line[CPBUF_SIZE_MAX];
	int len;

	len = SSL_read(src->t, line, CPBUF_SIZE_MAX);
	if (len > 0) {
		write(dst, line, len);
		fsync(dst);
		return 0;
	}
	else
		return 1;
}

void control(tunnel *t, int child, int cmd_in, int cmd_out)
{
	int len, ret;
	fd_set r_fds;
	struct timeval tv;
		
	FD_ZERO(&r_fds);
	
	for ( ; ; ) {
		FD_SET(cmd_out, &r_fds);
		memset(&tv, 0, sizeof(struct timeval));
		tv.tv_sec = select_to;
		
		ret = select(cmd_out + 1, &r_fds, NULL, NULL, &tv);
		if (ret == -1) {
			qerr("select failed");
			break;
		}
			/* copy message from server to stdin of command */
		if (len = SSL_pending(t->t)) {
			if (!copy_tun2fd(t, cmd_in)) {
				qerr("copy_tun2fd failed");
				break;
			}
		}
			/* copy stdout of command to tunnel */
		if (FD_ISSET(cmd_out, &r_fds)) {
			if ((ret = copy_fd2tun(cmd_out, t)) == SPTC_CMD_EOF) {
#ifdef DEBUG
				qerr("seen eof from command");
#endif
				break;
			}
			else if (ret) {
				tunnel_down(t);
			}
		}
		if ((t->state & UP) && ((time(NULL) - t->last_write) > inactivity_to)) {
#ifdef DEBUG
			qerr("shutting down inactive tunnel");
#endif
			tunnel_down(t);
		}
	}
}

#if !HAVE_DAEMON
	/* conditionally compiled */
	/* borrowed from stunnel */
void daemonize()
{
	int i;
	switch (fork()) {
		case -1:	/* error */
			qerr("fork failed");
			exit(1);
		case 0:	/* child */
			break;
		default:	/* parent */
			exit(0);
	}
	if (setsid() == -1) {
		qerr("setsid failed");
		exit(1);
	}
	for (i = 0; i < 3; i++)
		close(i);
}
#endif

void qerr(char *msg)
{
	syslog(SPTC_PRI, "%s", msg);
}

void dbl_close(int *x, int *y)
{
	close(x[0]);
	close(y[1]);
}

int process_options(int argc, char **argv)
{
	int option;
	struct passwd *user;
	struct group *group;
	
	while ((option = getopt(argc, argv, "c:p:t:i:w:u:g:v")) != -1) {
		switch (option) {
#if VERIFY_CERT
			case 'c':
				cert_path = optarg;
				break;
#endif
			case 'p':
				pid_file = optarg;
				break;
			case 't':
				select_to = atoi(optarg);
				break;
			case 'i':
				inactivity_to = atoi(optarg);
				break;
			case 'w':
				if (chdir(optarg) == -1) {
					qerr("chdir failed");
				}
				break;
			case 'u':
				user = getpwnam(optarg);
				if (!user)
					qerr("cannot get user info");
				sptc_uid = user->pw_uid;
				if (sptc_gid == -1)
					sptc_gid = user->pw_gid;
				if (user->pw_dir) {
					if (chdir(user->pw_dir) == -1)
						qerr("cannot chdir to user's home directory");
				}
				break;
			case 'g':
				group = getgrnam(optarg);
				if (!group)
					qerr("cannot get group info");
				sptc_gid = group->gr_gid;
				break;
			case 'v':
				fprintf(stderr, "sptc version %s\nCopyright 2002, Jason Royes\n", VERSION);
				fprintf(stderr, 
					"-- Compile-time defaults --\n"
					"inactivity timeout\t%d\n"
					"poll timeout\t\t%d\n"
#if VERIFY_CERT
					"cert path\t\t%s\n"
#endif
					"pid file\t\t%s\n"
					"queue name\t\t%s\n"
					"q-state name\t\t%s\n"
					"--\n", 
					INACTIVITY_TO, SELECT_TO, 
#if VERIFY_CERT
					SPTC_CERT_DIR,
#endif
					SPTC_PID_FILE, Q_NAME, Q_STATE_NAME);
			default:
				usage(argv[0]);
		}
	}
	return optind;
}

	/* setup signals with sigaction */
void setup_signals(void)
{
        struct  sigaction pipe;
	struct	sigaction quit;
		
        memset(&pipe, 0, sizeof(struct sigaction));
	memset(&quit, 0, sizeof(struct sigaction));
		
        pipe.sa_sigaction = pipe_bus;
	quit.sa_sigaction = catch_term;
	
        quit.sa_flags = pipe.sa_flags = SA_SIGINFO;
	
                /* install signal handlers for SIGPIPE SIGTERM & SIGINT */
        sigaction(SIGPIPE, &pipe, NULL);
        sigaction(SIGTERM, &quit, NULL);
        sigaction(SIGINT, &quit, NULL);
}

#define PORT		(host+1)
#define CMD		(host+2)
#define ARG_START	(host+3)

int main(int argc, char **argv)
{
	int n_args;
	int cmd_in[2], cmd_out[2];
	int ret;
	int host;
	char **args, cmd[CMD_SIZE_MAX];
	
	inactivity_to = INACTIVITY_TO;
	select_to = SELECT_TO;
	cert_path = SPTC_CERT_DIR;
	pid_file = SPTC_PID_FILE;
	
		/* host, port, command, cmd-args */
	host = process_options(argc, argv);	/* host index */
	
	if (select_to <= 0)
		select_to = SELECT_TO;
	if (inactivity_to <= 0)
		inactivity_to = INACTIVITY_TO;
	
	if ((argc - ARG_START) <= 0)
		usage(argv[0]);
	
		/* daemonize baby! */
#ifdef HAVE_DAEMON
	daemon(2, 2);	/* provided by OS */
#else
	daemonize();	/* manual daemonization */ 
#endif

	openlog(argv[0], LOG_PID, SPTC_FAC|SPTC_PRI);
	
		/* derive argv[0] (spawned cmd's perspective) from command line */
	cmd[CMD_SIZE_MAX] = 0;
	strncpy(cmd, basename(argv[CMD]), 
		CMD_SIZE_MAX - 1);				/* copy the command name */
	n_args = argc - ARG_START;				/* calc. the number of args being passed to command */
	args = (char **)malloc(sizeof(char *) * (argc + 1));	/* one extra for the NULL */
	if (!args) {
		qerr("out of memory");
		exit(E_SPTC_NOMEM);
	}

	args[0] = cmd;
	memcpy(&args[1], &argv[ARG_START], sizeof(char *) * argc);
	args[argc] = NULL;
	
		/* get pipes for duplex communication with spawned command */
	if (pipe(cmd_in) || pipe(cmd_out)) {
		qerr("cannot get pipe");
		exit(E_SPTC_GETPIPE);
	}
	
		/* initialize libssl and perform tunnel setup*/
	SSL_library_init();
	SSL_load_error_strings();
	
#ifdef RAND_SOURCE
        if (RAND_egd(RAND_SOURCE) == -1)
                qerr("RAND_egd failed");
#endif
	
		/* initialize SSL context */
	t_ctx = SSL_CTX_new(TLSv1_client_method());
	SSL_ECHK(t_ctx, "context", E_SPTC_TCTX);
	
		/* all SSL objects will verify certificates by checking this directory */
#if VERIFY_CERT
	SSL_CTX_set_verify(t_ctx, SSL_VERIFY_PEER, NULL);
	SSL_ECHK(SSL_CTX_load_verify_locations(t_ctx, NULL, cert_path), "verify location", E_SPTC_VFYLOC);
#endif
	
	if (!tunnel_init(t_ctx, &t, argv[host], atoi(argv[PORT]))) {
		qerr("cannot tunnel_init");
		return ret;
	}

	SSL_ECHK((tunnel_up(&t) != E_TUN_AUTH), "Authentication failed", E_TUN_AUTH);
	
		/* write the pid file */
	if (!write_pid(pid_file, getpid()))
		qerr("unable to write pid file");	
	if (sptc_uid > -1 && sptc_gid > -1) {
			/* chown pid file so we can remove it @ cleanup */
		if (chown(pid_file, sptc_uid, sptc_gid) == -1)
			qerr("cannot chown() pid file");
			/* set gid before uid so as not to kneecap ourselves */
		if (setgid(sptc_gid) == -1)
			qerr("cannot setgid()");
		if (setuid(sptc_uid) == -1)
			qerr("cannot setuid()");
	}
	
	child_pid = fork();
	if (child_pid == 0) {
		dbl_close(cmd_out, cmd_in);
		spawn_command(argv[CMD], args, cmd_in[0], cmd_out[1]);
		syslog(SPTC_PRI, "cannot spawn command %s", argv[3]);
		exit(E_SPTC_EXEC);
	}

	if (child_pid == -1) {
		qerr("unable to fork");
		cleanup();
	}
	
	setup_signals();

	dbl_close(cmd_in, cmd_out);	/* close pipe ends we don't need */
	
	syslog(SPTC_PRI, "v%s initialized, cmd=%s/host=%s/port=%s/s_to=%d/i_to=%d", 
		VERSION, argv[CMD], argv[host], argv[PORT], select_to, inactivity_to);

	control(&t, child_pid, cmd_in[1], cmd_out[0]);

	cleanup();
}
