#ifndef TUN_H
#define TUN_H

#if HAVE_NETINET_IN_H
# include <netinet/in.h>
#endif

#include <time.h>
#if HAVE_SYSLOG_H
# include <syslog.h>
#endif

#define TUN_PRI		LOG_ERR

	/* Fault tolerant tunnel API */
#define Q_STATE_NAME	"queue-state.sptc"
#define Q_NAME		"queue.sptc"

	/* 5 megs of queue */
#ifndef Q_SIZE_MAX
#define Q_SIZE_MAX	5242880
#endif

	/* try reopening the tunnel every N seconds */
#ifndef TUN_RETRY
#define TUN_RETRY	10
#endif

#define FLUSH_BUF_SIZE		4096
#define CONNECT_TO		60	/* retry tunnel if >= 60 seconds passed since last connect attempt */
#define UP			0x1
#define DOWN			0x2
#define FLUSHING		0x4

	/* Errors */
#define E_TUN_BASE		300
#define E_TUN_QFULL		E_TUN_BASE+1
#define E_TUN_QNOTOPEN		E_TUN_BASE+2
#define E_TUN_QWRITE		E_TUN_BASE+3
#define E_TUN_WRITE		E_TUN_BASE+4
#define E_TUN_BADTNL		E_TUN_BASE+5
#define E_TUN_SETFD		E_TUN_BASE+6
#define E_TUN_CONNECT		E_TUN_BASE+7
#define E_TUN_AUTH		E_TUN_BASE+8
#define E_TUN_STOPFLUSH		E_TUN_BASE+9
#define E_TUN_QISOPEN		E_TUN_BASE+10
#define E_TUN_QCORRUPT		E_TUN_BASE+11
#define E_TUN_BADSEEK		E_TUN_BASE+12
#define E_TUN_NOQFILE		E_TUN_BASE+13
#define E_TUN_QSTATE		E_TUN_BASE+14
#define E_TUN_WRSTATE		E_TUN_BASE+15
#define E_TUN_QCLOSED		E_TUN_BASE+16

	/* Queue attributes */	
typedef struct _queue {
	char *q_name;
	char *q_state_name;
	int q_fd;
	unsigned int size;
} queue;
	
	/* Tunnel attributes */
typedef struct _tunnel {
	char *host;
	uint16_t port;
	SSL *t;
	char state;		/* UP/DOWN | FLUSHING */
	queue q;
	int s;			/* socket */
	time_t last_connect;	/* time of last try to bring tunnel online: to prevent SYN floods */
	time_t last_write;	/* time of last write to tunnel */
} tunnel;

int tunnel_up(tunnel *t);
int tunnel_down(tunnel *t);
int tunnel_init(SSL_CTX *ctx, tunnel *t, char *host, uint16_t port);
int tunnel_write(tunnel *t, char *data, int len);
int tunnel_flush(tunnel *t);
int queue_close(queue *q);
int queue_get_size(queue *q);
int queue_open(queue *q, int mode);
int queue_write(queue *q, char *data, int len);

#endif
